/* VRANK pattern matching suite
 * Copyright (C) 1996 to Justin Zobel
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * See Copyright.txt for further details. */


/* Code for maintaining a splay tree of grams. */
/* Currently only one tree at a time is possible. */

#include "def.h"

static TNODE *nodecreate(char *, TNODE *);

static TNODE *tree;

void
inittree()
{
    tree = NULL;
}

/* Perform one left or one right rotation to bring current node one level
   closer to root */
#define ROTATE(wt) \
	{ \
	    register TNODE	*tp, *tgp; \
	\
	    tp = wt->par; tgp = wt->par->par; \
	    wt->par = tgp; \
	    if( tgp != NULL ) \
		if( tgp->left == tp )	tgp->left = wt; \
		else			tgp->right = wt; \
	    tp->par = wt; \
	    if( wt == tp->left ) \
	    { \
		if( wt->right != NULL ) \
		    wt->right->par = tp; \
		tp->left = wt->right; \
		wt->right = tp; \
	    } else { \
		if( wt->left != NULL ) \
		    wt->left->par = tp; \
		tp->right = wt->left; \
		wt->left = tp; \
	    } \
	}

#define ROTATEFAC 4
int	scount = 0;

/* Insert gram into a splay tree.  If gram is already present, bring it to
   root (sometimes).  Always return it. */
TNODE *
tinsert(char *gram)
{
    TNODE	*curr = tree, *par, *gpar, *prev = NULL;
    int		val;

    scount = (scount+1) % ROTATEFAC;

    if( curr == NULL )
    {
	tree = nodecreate(gram, NULL);
	return(tree);
    }
    while( curr != NULL && ( val = strncmp(gram, curr->gram, GRAMLEN) ) != 0 )
    {
	prev = curr;
	if( val > 0 )    curr = curr->right;
	else	curr = curr->left;
    }
    if( curr == NULL )
    {
	if( val > 0 )    curr = prev->right = nodecreate(gram, prev);
	else	curr = prev->left = nodecreate(gram, prev);
    }

    if( scount == 0 )    /* Move node to root */
    {
	while( (par = curr->par) != NULL )
	    if( (gpar = par->par) == NULL )
	    {
		ROTATE(curr)
	    }
	    else if( (par->left == curr && gpar->left == par) ||
		(par->right == curr && gpar->right == par) )
	    {
		ROTATE(par)
		ROTATE(curr)
	    }
	    else
	    {
		ROTATE(curr)
		ROTATE(curr)
	    }
	tree = curr;
    }

    return(curr);
}


/* Fetch gram from a splay tree.  If gram is present, bring it to
   root (sometimes).  Always return it. */
TNODE *
tfetch(char *gram)
{
    TNODE	*curr = tree, *par, *gpar, *prev = NULL;
    int		val;

    scount = (scount+1) % ROTATEFAC;

    while( curr != NULL && (val = strncmp(gram, curr->gram, GRAMLEN)) != 0 )
    {
	prev = curr;
	if( val > 0 )    curr = curr->right;
	else	curr = curr->left;
    }
    if( curr == NULL )
	return(NULL);

    if( scount == 0 )    /* Move node to root */
    {
	while( (par = curr->par) != NULL )
	    if( (gpar = par->par) == NULL )
	    {
		ROTATE(curr)
	    }
	    else if( (par->left == curr && gpar->left == par) ||
		(par->right == curr && gpar->right == par) )
	    {
		ROTATE(par)
		ROTATE(curr)
	    }
	    else
	    {
		ROTATE(curr)
		ROTATE(curr)
	    }
	tree = curr;
    }

    return(curr);
}


/* Create a node to hold a gram */
TNODE *
nodecreate(char *gram, TNODE *par)
{
    TNODE    *tmp;

    tmp = (TNODE *) malloc(sizeof(TNODE));
    strncpy(tmp->gram, gram, GRAMLEN);
    tmp->left = tmp->right = NULL;
    tmp->par = par;
    initvec(&tmp->vec);

    return(tmp);
}


BITVEC *
gettvec(TNODE *tree)
{
    return(&tree->vec);
}


void
cleanuptree()
{
    void	docleanuptree(TNODE *);

    docleanuptree(tree);
}


void
docleanuptree(TNODE *thetree)
{
    if( thetree != NULL )
    {
	docleanuptree(thetree->left);
	padvec(&thetree->vec);
	docleanuptree(thetree->right);
    }
}


/* To find out size of ptr lists */
int
ptrsize()
{
    int		doptrsize(TNODE *);

    return(doptrsize(tree));
}


int
doptrsize(TNODE *t)
{
    return(( t==NULL ) ? 0 : 1 + veclen(&t->vec)/8
			    + doptrsize(t->left) + doptrsize(t->right));
}


/* To find out number of grams */
int
treesize()
{
    int		dotreesize(TNODE *);

    return(dotreesize(tree));
}


int
dotreesize(TNODE *t)
{
    return( ( t == NULL ) ? 0 : 1 + dotreesize(t->left) + dotreesize(t->right));
}


/* Assumes that tree has been reset -- could go berserk if not */
void
dumptree(char *fname)
{
    void	dodumptree(TNODE *,FILE *);
    FILE	*fp, *fopen();

    if( ( fp = fopen(fname, "w") ) == NULL )
    {
	fprintf(stderr, "Can't open %s for writing.\n", fname);
	exit(1);
    }
    dodumptree(tree, fp);
    fclose(fp);
}

void
dodumptree(TNODE *thetree, FILE *fp)
{
    if( thetree != NULL )
    {
	dodumptree(thetree->left, fp);
	fwrite(thetree->gram, sizeof(char), GRAMLEN, fp);
#ifdef DEBUG
	fprintf(stderr, "%c%c ",thetree->gram[0], thetree->gram[1]);
#endif /* DEBUG */
	dumpvec(&thetree->vec, fp);
#ifdef DEBUG
	fprintf(stderr, "#\n");
#endif /* DEBUG */
	dodumptree(thetree->right, fp);
    }
}
