/* VRANK pattern matching suite
 * Copyright (C) 1996 to Justin Zobel
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * See Copyright.txt for further details. */


/* Generate an index for vocab and dump it to vocindex.
   Assumes that vocab will be numbered from 1 to i */
/* Usage is mkvind [-f bfac] [-i index] [-v vocab] [-r] [-s type] [-h thresh]
    -r: assume range optimisation will be used, don't index first gram
    -s type: perform phonetic transformation of type before indexing
    -f bfac: override default blocking factor (BLOCKING)
    -i index: override default suffix (INDSUF)
    -v vocab: override default vocabulary (VOCAB)
    -h thresh: override default threshold (THRESH)

   Legal strings are of alnumerics and #
*/
/* There are two kinds of indexing.  In the ordinary mode, there is an array
   of ptrs to strings, and the gram pointer lists refer to this array.
   In DIRECT indexing, the gram pointers are directly to the strings.
   (Querying not implemented for direct indexing.)
*/

#include "def.h"

int	blocking = BLOCKING;

main(int argc, char *argv[])
{
    int		ftime(), atoi(), getrusage();
    double	ceil(), log10();
    int		i, j, k;
    int		len;			/* length of cur. word, inc. ^, $ */
    int		threshold = THRESHOLD;
    int		dorange = false;
    int		dophonetic = false;	/* type of phonetic */
    int		ng;			/* number of distinct grams */
    char	buf[BUFLEN], *ptr, gram[GRAMLEN], *fgets();
    char	*vocab = VOCAB, *indsuf = INDSUF, *paramsuf = PARAMSUF;
    char	buf2[BUFLEN], vocindex[BUFLEN], vocparam[BUFLEN];
    char	*skipsuf = SKIPSUF, vocsindex[BUFLEN];
    FILE	*fp, *fopen();
    TNODE	*thenode;
    int		base;			/* posn of start of current word */
    double	seconds;
#ifdef SOLARIS
    struct tms	tms_was, tms_now;
#else /* SOLARIS */
    struct rusage ru_was, ru_now;
#endif /* SOLARIS */

restart:
    if( argc >= 2 && strcmp(argv[1], "-r") == 0 )
    {
	dorange = true;
	argc--, argv++;
	goto restart;
    }
    if( argc >= 3 && strcmp(argv[1], "-h") == 0 )
    {
	threshold = atoi(argv[2]);
	argc -= 2, argv += 2;
	goto restart;
    }
    if( argc >= 3 && strcmp(argv[1], "-f") == 0 )
    {
	blocking = atoi(argv[2]);
	argc -= 2, argv += 2;
	goto restart;
    }
    if( argc >= 3 && strcmp(argv[1], "-i") == 0 )
    {
	indsuf = argv[2];
	argc -= 2, argv += 2;
	goto restart;
    }
    if( argc >= 3 && strcmp(argv[1], "-s") == 0 )
    {
	dophonetic = atoi(argv[2]);
	if( phonetictype(dophonetic) == NULL )
	{
	    fprintf(stderr, "Unknown phonetic type %d.\n", dophonetic);
	    exit(1);
	}
	argc -= 2, argv += 2;
	goto restart;
    }
    if( argc >= 3 && strcmp(argv[1], "-v") == 0 )
    {
	vocab = argv[2];
	argc -= 2, argv += 2;
	goto restart;
    }
    if( argc >= 2 )
    {
	printf("Usage: mkvind [-f blockfac] [-h threshold] [-i index]\n");
	printf("              [-s type] [-v vocab]\n");
	exit(1);
    }

#if DIRECT
    if( blocking > 1 )
    {
	fprintf(stderr, "Blocking not implemented for direct access.\n");
	exit(1);
    }
#endif /* DIRECT */

    sprintf(vocindex, "%s.%s", vocab, indsuf);
    sprintf(vocsindex, "%s.%s.%s", vocab, indsuf, skipsuf);
    sprintf(vocparam, "%s.%s.%s", vocab, indsuf, paramsuf);

    unlink(vocindex);
    unlink(vocsindex);
    unlink(vocparam);

    initcmp();
    inittree();

    if( ( fp = fopen(vocab, "r") ) == NULL )
    {
	fprintf(stderr, "Can't open %s.\n", vocab);
	exit(1);
    }

    /* Hack to get start/end of line matching, if gramlen >= 2 */
    if( GRAMLEN > 1 && !dorange )
    /* by magic this even works when blocking is used!  The binary
       search will bracket strings starting with the appropriate ngram */
    {
	ptr = buf + 1;
        buf[0] = '^';
    }
    else
	ptr = buf;

	/* Start timing */
#ifdef SOLARIS
    (void) times(&tms_was);
#else /* SOLARIS */
    (void) getrusage(RUSAGE_SELF, &ru_was);
#endif /* SOLARIS */

    for( base=0, i=0 ; i<MAXVOC && fgets(buf2, BUFLEN, fp) != NULL ; i++ )
    {
	if( dophonetic )
	    phonetic(dophonetic, buf2, strlen(buf2), ptr);
	else
	    strcpy(ptr, buf2);
	len = strlen(ptr);
	if( ptr[len-1] == '\n' )
	{
	    ptr[len-1] = '\0';
	    len--;
	}
	for( j=0 ; j<len ; j++ )
	    if( !validchar(ptr[j]) )
	    {
		fprintf(stderr, "Bad vocab entry: %s\n", ptr);
		exit(1);
	    }
	if( GRAMLEN > 1 )
	{
	    len += 2;
	    buf[len-1] = '$';
	    buf[len] = '\0';
	}

	for( j=0 ; j<len+1-GRAMLEN ; j++ )
	{
	    for( k=0 ; k<GRAMLEN ; k++ )
		gram[k] = buf[j+k];
	    thenode = tinsert(gram);
#if DIRECT
	    putrunlen(gettvec(thenode), base);
#else
	    putrunlen(gettvec(thenode), i/blocking);
#endif /* DIRECT */
	}
	base += ( GRAMLEN > 1 ) ? len : len + 1;
    }

    if( i >= MAXVOC )
    {
	fprintf(stderr, "Vocabulary too big.\n");
	exit(1);
    }

    fclose(fp);
    cleanuptree();
    ng=treesize();
    dumpparams(vocparam, i, ng, blocking, dophonetic, threshold);
    dumptree(vocindex);
#ifdef SOLARIS
    (void) times(&tms_now);
    seconds = ( ( tms_now.tms_utime + tms_now.tms_stime )
	    - ( tms_was.tms_utime + tms_was.tms_stime ) ) / (double) CLK_TCK;
#else /* SOLARIS */
    (void) getrusage(RUSAGE_SELF, &ru_now);
    seconds = ( ru_now.ru_utime.tv_sec - ru_was.ru_utime.tv_sec )
	    + 0.000001*( ru_now.ru_utime.tv_usec - ru_was.ru_utime.tv_usec );
#endif /* SOLARIS */

    /*ng = treesize();*/
    i--;
#if BRIEF
    printf("#N,r\tvoclen\tvocsize\tbfac\tvecs\tacc\tptrs\tthresh\tcpu  #mkvind\n");
#if DIRECT
    printf("%d %c\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%1.3f  #mkvind\n",
	GRAMLEN, (dorange?'y':'n'), i+1, base, blocking, ng,
	( 1 + GRAMLEN*8*ng /* grams */
	    + ng*32	   /* ptrs to index entries */
	) / 8,
	ptrsize(), threshold, seconds);
#else /* DIRECT */
    printf("%d %c\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%1.3f  #mkvind\n",
	GRAMLEN, (dorange?'y':'n'), i+1, base, blocking, ng,
	( 1 + GRAMLEN*8*ng /* grams */
	    + ng*32	   /* ptrs to index entries */
			/* ptrs to words */
	    + ( 1+i/blocking ) * (int)(ceil(L2((double)base)) + 0.01)
	) / 8,
	ptrsize(), threshold, seconds);
#endif /* DIRECT */
#else /* BRIEF */
    printf("Found %d vocabulary entries and %d %d-grams, range opt %s\n",
	--i, ng, GRAMLEN, (dorange?"on":"off"));
    printf("Vocab size is %d bytes (incl. nulls).\n", base);
#if DIRECT
    printf("Est. access structure size is %d bytes (excluding ptr lists).\n",
	( 1 + GRAMLEN*8*ng /* grams */
	    + ng*32	   /* ptrs to index entries */
	) / 8
    );
#else /* DIRECT */
    printf("Est. access structure size is %d bytes (excluding ptr lists).\n",
	( 1 + GRAMLEN*8*ng /* grams */
	    + ng*32	   /* ptrs to index entries */
			/* ptrs to words */
	    + ( 1+i/blocking ) * (int)(ceil(L2((double)base)) + 0.01)
	) / 8
    );
#endif /* DIRECT */
    printf("Est. size of ptr lists is %d bytes (blocking factor %d)\n",
	ptrsize(), blocking
    );
    printf("Used %1.3f cpu sec.\n", seconds);
#endif /* BRIEF */

    exit(0);
}


/* Dump all the index and vocabulary parameters */
void
dumpparams(char *fname, int vocsize, int ng, int bfac, int dp, int thresh)
{
    int		glen = GRAMLEN;
    FILE 	*fp, *fopen();

    if( ( fp = fopen(fname, "w") ) == NULL )
    {
	    fprintf(stderr, "Can't open %s for writing.\n", fname);
	    exit(1);
    }
    fwrite(&vocsize, sizeof(int), 1, fp);
    fwrite(&ng, sizeof(int), 1, fp);
    fwrite(&glen, sizeof(int), 1, fp);
    fwrite(&bfac, sizeof(int), 1, fp);
    fwrite(&dp, sizeof(int), 1, fp);
    fwrite(&thresh, sizeof(int), 1, fp);
    fclose(fp);
}
