/* VRANK pattern matching suite
 * Copyright (C) 1996 to Justin Zobel
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * See Copyright.txt for further details. */


/* Phonix algorithm, roughly as described by Gadd.  Doesn't have the
   second (numerical coding) phase. */

/* J. Zobel, June 1993 */

#include <stdio.h>
#include <ctype.h>

    /* Optional: add 'y' as vowel */
#define ISVOWEL(c)	( (c)=='a'||(c)=='e'||(c)=='i'||(c)=='o'||(c)=='u' )
#define ISCONS(c)	( islower(c) && ! ISVOWEL(c) )

#define true		1
#define false		0

#define PHONYMS		"/home/phobos/staff/jz/work/soundex/phonix-table"

#define LETCOUNT	28
#define VOWELPOS	26
#define CONSPOS		27
#define STRLEN		100

#define STARTS		rules[0]
#define MIDS		rules[1]
#define ENDS		rules[2]

typedef struct rulerec
{
    char	*in, *out;
    int		inlen, outlen;
    struct rulerec *next;
} RULE;

static RULE *rules[3][LETCOUNT];

void	loadphontable();
void	phonix(char *, int, char *);
void	ntphonix(char *, int, char *);
void	partphonix(char *, int, char *);

static void transform(char *, int, char *);
static RULE *newrule(RULE *, char *);
static void phoncpy(char *, char *, int, char *, int);
static int phoncmp(char *, char *, int);

char	*malloc(int);
char	*strncpy(char *, char *, int);

int	loaded = false;


/* Phonix algorithm */

/* Standard mapping:
   aehiouwy -> 0, which is omitted
   bp -> 1
   cgjkq -> 2
   dt -> 3
   l -> 4
   mn -> 5
   r ->6
   fv -> 7
   sxz -> 8
*/
/*                abcdefghijklmnopqrstuvwxyz */
#define PMAPPING "01230720022455012683070808"

void
phonix(char *in, int inlen, char *out)
{
    char	buf[STRLEN];
    int		c, lc, prev;
    int		i;

    if( !loaded )
	loadphontable();

    transform(in, inlen, buf);
    inlen = strlen(buf);
    in = buf;

    strcpy(out, "a000");
    prev = '0';

    for( i=0 ; inlen > 0 && i < 4 ; in++, inlen-- )
    {
	if( isalpha(*in) )
	{
	    lc = isupper(*in) ? tolower(*in) : *in;
	    c = PMAPPING[lc-'a'];
	    if ( i == 0 || ( c != '0' && c != prev ) )
	    {
		out[i] = i ? c : lc;
		i++;
	    }
	    prev = c;
	}
    }
    /* out[i] = '\0'; */
}


/* Nontruncating phonix */
void
ntphonix(char *in, int inlen, char *out)
{
    char	buf[STRLEN];
    int		c, lc, prev;
    int		i;

    if( !loaded )
	loadphontable();

    transform(in, inlen, buf);
    inlen = strlen(buf);
    in = buf;

    prev = '0';

    for( i=0 ; inlen > 0 ; in++, inlen-- )
    {
	if( isalpha(*in) )
	{
	    lc = isupper(*in) ? tolower(*in) : *in;
	    c = PMAPPING[lc-'a'];
	    if ( i == 0 || ( c != '0' && c != prev ) )
	    {
		out[i] = i ? c : lc;
		i++;
	    }
	    prev = c;
	}
    }
    out[i] = '\0';
}


void
partphonix(char *in, int inlen, char *out)
{
    int		i, j;

    if( !loaded )
	loadphontable();

    transform(in, inlen, out);

	/* now get rid of double letters */
	/* should also get rid of vowels? */
    for( i=j=1 ; out[i] != '\0' ; i++ )
	if( out[i] != out[j-1] )
	{
	    out[j] = out[i];
	    j++;
	}
    out[j] = '\0';
}


/* Each entry in table is of the form `OLD\tNEW'.  Index by first char
   if start or mid, by last char if end */
/* Lists are built in reverse, to ensure that subsuming rules come first */
void
loadphontable()
{
    int		i, pos;
    FILE	*fp, *fopen();
    char	*fgets(), buf[STRLEN];
    RULE	*ptr;

    loaded = true;

    if( ( fp = fopen(PHONYMS, "r") ) == NULL )
    {
	fprintf(stderr, "Error: can't open %s\n", PHONYMS);
	exit(1);
    }

    for( i=0 ; i<LETCOUNT ; i++ )
	STARTS[i] = MIDS[i] = ENDS[i] = NULL;

    fgets(buf, STRLEN, fp);	/* buf = "#start" */

    for( fgets(buf, STRLEN, fp) ; *buf != '#' ; fgets(buf, STRLEN, fp) )
	/* read a block of start rules */
    {
	if( *buf == 'V' )
	    pos = VOWELPOS;
	else if( *buf == 'C' )
	    pos = CONSPOS;
	else
	    pos = *buf - 'a';

	STARTS[pos] = newrule(STARTS[pos], buf);
    }
    for( fgets(buf, STRLEN, fp) ; *buf != '#' ; fgets(buf, STRLEN, fp) )
	/* read a block of mid rules */
    {
	if( *buf == 'V' )
	    pos = VOWELPOS;
	else if( *buf == 'C' )
	    pos = CONSPOS;
	else
	    pos = *buf - 'a';

	MIDS[pos] = newrule(MIDS[pos], buf);
    }
    for( fgets(buf, STRLEN, fp) ; *buf != '#' ; fgets(buf, STRLEN, fp) )
	/* read a block of end rules */
    {
	for( i=0 ; buf[i] != '\t' ; i++ )
	    ;
	i--;
	if( buf[i] == 'V' )
	    pos = VOWELPOS;
	else if( buf[i] == 'C' )
	    pos = CONSPOS;
	else
	    pos = buf[i] - 'a';

	ENDS[pos] = newrule(ENDS[pos], buf);
    }

/* #define DEBUG */
#ifdef DEBUG
    for( i=0 ; i<3 ; i++ )
	for( pos=0 ; pos<LETCOUNT ; pos++ )
	{
	    printf(">> %d, %c <<", i, pos+'a');
	    for( ptr=rules[i][pos] ; ptr!=NULL ; ptr=ptr->next )
		printf("--%s %s\n", ptr->in, ptr->out);
	}
#endif /* DEBUG */

}


/* Turn line of input into a rule */
RULE *
newrule(RULE *next, char *in)
{
    int		i;
    RULE	*ptr;
    char	buf[STRLEN];

    ptr = (RULE *) malloc(sizeof(RULE));
    ptr->next = next;

	    /* get `in' form */
    for( i=0 ; *in != '\t' ; i++, in++ )
	buf[i] = *in;
    buf[i] = '\0';
    ptr->in = malloc(i+1);
    strcpy(ptr->in, buf);
    ptr->inlen = i;

    for( ; *in == '\t' ; in++ )
	;

	    /* get `out' form */
    for( i=0 ; *in != '\n' ; i++, in++ )
	buf[i] = *in;
    buf[i] = '\0';
    if( buf[0] == '%' )	/* null out */
	buf[0] = '\0';
    ptr->out = malloc(i+1);
    strcpy(ptr->out, buf);
    ptr->outlen = i;

    return(ptr);
}


/* Perform phonetic transformation */
void
transform(char *orig, int inlen, char *out)
{
    char	tail[STRLEN];			/* the suffix of out */
    char	in[STRLEN];			/* buffer for orig word */
    RULE	*ptr;
    int		c;				/* index char into rules */
    int		pos;				/* index into rules */
    int		i;				/* num of orig chars consumed */
    int		j;				/* num of chars in out */
    int		fd;				/* boolean */

    for( i=0 ; i<inlen ; i++ )
	in[i] = isupper(orig[i]) ? tolower(orig[i]) : orig[i];

    i = j = 0;
    tail[0] = '\0';

#define FINDTEST(off) \
		{ \
		    if( i+ptr->inlen <= inlen && \
			    phoncmp(ptr->in, in+off, ptr->inlen)==0 ) \
			fd = true; \
		    else \
			ptr = ptr->next; \
		}

#define RUNTESTS(off,type) \
		{ \
		    fd = false; \
		    if( ISVOWEL(c) ) \
			for( ptr=type[VOWELPOS] ; ptr!=NULL && !fd ; ) \
			    FINDTEST(off) \
		    else if( ISCONS(c) ) \
			for( ptr=type[CONSPOS] ; ptr!=NULL && !fd ; ) \
			    FINDTEST(off) \
		    if( !fd && islower(c) ) \
			for( ptr=type[c-'a'] ; ptr!=NULL && !fd ; ) \
			    FINDTEST(off) \
		}

	/* Process `start's */
    c = in[i];
    RUNTESTS(i,STARTS)

    if( fd )
    {
#ifdef DEBUG
	printf("(%s:%d,%s:%d),%d\n",ptr->in,ptr->inlen,ptr->out,ptr->outlen,0);
#endif /* DEBUG */
	phoncpy(out+j, ptr->out, ptr->outlen, in+i, ptr->inlen);
	i += ptr->inlen;
	j += ptr->outlen;
    }
    else
    {
	out[j] = in[i];
	i++;
	j++;
    }


	/* Process `end's */
    fd = false;
    c = in[inlen-1];
    RUNTESTS(inlen-ptr->inlen,ENDS)

    if( fd )
    {
#ifdef DEBUG
	printf("(%s:%d,%s:%d),%d\n",ptr->in,ptr->inlen,ptr->out,ptr->outlen,2);
#endif /* DEBUG */
	phoncpy(tail, ptr->out, ptr->outlen, in+inlen-ptr->inlen, ptr->inlen);
	tail[ptr->outlen] = '\0';
	inlen -= ptr->inlen;
    }
    else
    {
	tail[0] = c;
	tail[1] = '\0';
	inlen--;
    }


	/* Process `mid's */
    while( i < inlen )
    {
	fd = false;
	c = in[i];
	RUNTESTS(i,MIDS)

	if( fd )
	{
#ifdef DEBUG
	printf("(%s:%d,%s:%d),%d\n",ptr->in,ptr->inlen,ptr->out,ptr->outlen,1);
#endif /* DEBUG */
	    phoncpy(out+j, ptr->out, ptr->outlen, in+i, ptr->inlen);
	    i += ptr->inlen;
	    j += ptr->outlen;
	}
	else
	{
	    out[j] = in[i];
	    i++;
	    j++;
	}
    }

    out[j] ='\0';
    strcat(out, tail);
}


/* like strcmp, but pat can contain C and V */
int
phoncmp(char *pat, char *s, int len)
{
    for( ; len>0 ; len--, pat++, s++ )
    {
	if( *pat == 'C' )
	{
	    if( !ISCONS(*s) )
		return(1);
	}
	else if( *pat == 'V' )
	{
	    if( !ISVOWEL(*s) )
		return(1);
	}
	else if( *pat != *s )
	    return( *s - *pat );
    }

    return(0);
}


/* like strncpy, but uses pat and orig to fix up C, V */
/* C, V only handled at end of strings */
void
phoncpy(char *out, char *in, int len, char *orig, int origlen)
{
    strncpy(out, in, len);
    if( out[0] == 'C' || out[0] == 'V' )
	out[0] = orig[0];
    if( out[len-1] == 'C' || out[len-1] == 'V' )
	out[len-1] = orig[origlen-1];
}
