/* VRANK pattern matching suite
 * Copyright (C) 1996 to Justin Zobel
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * See Copyright.txt for further details. */


/* Query front end for finding words in vocabulary. */
/* Usage is vfind [-b] [-r] [-s] [-t] [-rpt num]
		  [-h threshold] [-i index] [-v vocab] [-skip]
    -b: use brute force search to find answers
    -r: use binary search to narrow down search range; ASSUMES sorted
	vocabulary
    -s: silent; show final stats only;
    -t: show time only; don't show answers (allows retrieval without i/o)
    -rpt num: do each query num times over, to improve accuracy of timing
    -i index: override default index (INDSUF)
    -h threshold: override default threshold (in params file)
    -v vocab: override default vocabulary (VOCAB)
    -skip: To activate the use of skip in inverted file
*/

#include "def.h"

int	bfac;					/* blocking factor */
int	dophonetic;				/* whether index is phonetic */
int	dorange = false;			/* range optimisation */
int	skip = false;   			/* No skip used initially*/
int	threshold, origthresh = -1;		/* f.m. threshold */
int	r_low, r_high;				/* range for searching */
int	*wdnums;				/* matching words */
char	**voc;					/* the vocabulary */
int	vocsize;				/* Vocabulary size */
int     noofgrams;				/* Number of grams */
int 	glen;					/* Gram length */
int	ilproc = 0;				/* Inverted lists processed */
    /* Following is used to simulate storage of vocab on disk */
int	*diskblocks;				/* Array of block numbers */	
int	blockacc = 0;				/* Number of blocks accessed */

static void loadvoc(char *, char **);
static int  findgrams(char *,char [][]);
static int  findmatches(char [][],int,int *);
static int  findsmatches(char [][],int,int *);
static int  loadindex(char *);
static int  loadsindex(char *);
static void findrange(char *, char *[]);
static void readparams(char *);
static void freemem(void);

main(int argc, char *argv[])
{
    int		off, ng, i, j, k, len, qcnt, getrusage(), ftime();
    int		brute = false;			/* brute force search */
    int		timeonly = false;		/* don't show answers */
    int		silent = false;			/* final status only */
    int		manytimes = false;		/* execute each q rpt times */
    int		numgrams, numwds;		/* numbers of things */
    int		tfc, ttc, fc, tc;		/* numbers of matches */
    int		brutes;				/* number of full searches */
    int		rpt = 1;			/* tot num of iterations */
    int		repeat;				/* number of iterations left */
    int		prevblock;			/* last disk block accessed */
    char	buf[BUFLEN], grams[BUFLEN][GRAMLEN], *strchr();
    char	*vocab = VOCAB, *indsuf = INDSUF,vocindex[BUFLEN];
    char 	*paramsuf = PARAMSUF, paramindex[BUFLEN];
    char 	*skipsuf=SKIPSUF;
    double	milli;
#ifdef SLOWMATCH
    int		ss, scnt;			/* for counting iterations */
#endif /* SLOWMATCH */
#ifdef SOLARIS
    char	*regcmp(const char *, ...);
    char	*regex(const char *, const char *, ...), *pat;
    struct tms	tms_was, tms_now;
#else /* SOLARIS */
    char	*re_comp(), *re_exec();
    struct rusage ru_was, ru_now;
#endif /* SOLARIS */

#if DIRECT
    fprintf(stderr, "Querying not implemented for direct indexes.\n");
    exit(1);
#endif /* DIRECT */

restart:
    if( argc >= 3 && strcmp(argv[1], "-h") == 0 )
    {
	origthresh = atoi(argv[2]);
	argc -= 2, argv += 2;
	goto restart;
    }
    if( argc >= 3 && strcmp(argv[1], "-i") == 0 )
    {
	indsuf = argv[2];
	argc -= 2, argv += 2;
	goto restart;
    }
    if( argc >= 3 && strcmp(argv[1], "-v") == 0 )
    {
	vocab = argv[2];
	argc -= 2, argv += 2;
	goto restart;
    }
    if( argc >= 3 && strcmp(argv[1], "-rpt") == 0 )
    {
	rpt = atoi(argv[2]);
	manytimes = true;
	argc -= 2, argv += 2;
	goto restart;
    }
    if( argc >= 2 && strcmp(argv[1], "-b") == 0 )
    {
	brute = true;
	argc--, argv++;
	goto restart;
    }
    if( argc >= 2 && strcmp(argv[1], "-r") == 0 )
    {
	dorange = true;
	argc--, argv++;
	goto restart;
    }
    if( argc >= 2 && strcmp(argv[1], "-s") == 0 )
    {
	silent = true;
	argc--, argv++;
	goto restart;
    }
    if( argc >= 2 && strcmp(argv[1], "-t") == 0 )
    {
	timeonly = true;
	argc--, argv++;
	goto restart;
    }
    if( argc >= 2 && strcmp(argv[1], "-skip") == 0 )
    {
	skip = true;
	argc--, argv++;
	goto restart;
    }

    if( argc >= 2 )
    {
	fprintf(stderr,"Usage: vfind [-b] [-r] [-s] [-skip] [-t] [-rpt num]\n");
	fprintf(stderr,"             [-h threshold] [-i index] [-v vocab]\n");
	exit(1);
    }

    sprintf(paramindex, "%s.%s.%s", vocab, indsuf, paramsuf);
    readparams(paramindex);	/* sets origthresh if not set */

    voc = (char **) malloc(vocsize*sizeof(char*));
    diskblocks = (int *) malloc(vocsize*sizeof(int));
    wdnums = (int *) malloc(QUERYMAX * sizeof(int));
    if( skip )
    {
        sprintf(vocindex, "%s.%s.%s", vocab, indsuf, skipsuf);
        sinithash(noofgrams);
    }
    else
    {
        sprintf(vocindex, "%s.%s", vocab, indsuf);
        inithash(noofgrams);
    }

#ifdef SOLARIS
    (void) times(&tms_was);
#else /* SOLARIS */
    (void) getrusage(RUSAGE_SELF, &ru_was);
#endif /* SOLARIS */

    loadvoc(vocab, voc);
		
    if( brute )
    {
	ng = 0;
	threshold = vocsize + 1;		/* not actually necessary */
	bfac = 1;
	dophonetic = false;
    }
    else
    {
	ng = (skip) ? loadsindex(vocindex) : loadindex(vocindex);
	    /* sets bfac, phonetic */
	    /* threshold is number of words, not number of blocks */
	threshold = 1 + ( origthresh - 1 )/bfac;
    }

#ifdef SOLARIS
    (void) times(&tms_now);
    milli = 1000.0 * ( ( tms_now.tms_utime + tms_now.tms_stime )
	    - ( tms_was.tms_utime + tms_was.tms_stime ) ) / (double) CLK_TCK;
#else /* SOLARIS */
    (void) getrusage(RUSAGE_SELF, &ru_now);
    milli = 1000.0 * ( ru_now.ru_utime.tv_sec - ru_was.ru_utime.tv_sec )
	+ 0.001*( ru_now.ru_utime.tv_usec - ru_was.ru_utime.tv_usec );
#endif /* SOLARIS */

    if( dophonetic )
    {
	fprintf(stderr, "Error: index is for phonetic queries only.\n");
	exit(1);
    }

#if !BRIEF
    printf("Startup took %1.3f cpu sec.\n", milli/1000.0);
#endif /* BRIEF */

	/* Get a string, parse into grams, merge ptr lists of
	   grams, print out all matching words. */
	/* Naive: doesn't go for shortest ptr list, etc. */
    if( !silent )
	printf("vf> ");
#ifdef SOLARIS
    (void) times(&tms_was);
#else /* SOLARIS */
    (void) getrusage(RUSAGE_SELF, &ru_was);
#endif /* SOLARIS */
    for( brutes=ttc=tfc=qcnt=0 ; gets(buf) != NULL ; qcnt++ )
    {
	fc = tc = 0;
	len = strlen(buf);
	if( len == 0 )
	    goto skiplook;

		/* Validate expression */
#ifdef SOLARIS
	if( ( pat = regcmp(buf, (char*)0) ) == NULL )
#else /* SOLARIS */
	if( re_comp(buf) != NULL )
#endif /* SOLARIS */
	{
	    if( !silent )
		printf("Wierd expression.\n");
	    goto skiplook;
	}
	if( ( strchr("^.", buf[0]) == NULL && !validchar(buf[0]) )
	    || ( strchr("$.*", buf[len-1]) == NULL
		 && !validchar(buf[len-1]) ) )
	{
	    if( !silent )
		printf("Wierd expression.\n");
	    goto skiplook;
	}
	for( i=1 ; i<len-1 ; i++ )
	{
	    if( strchr(".*", buf[i]) == NULL && !validchar(buf[i]) )
	    {
		if( !silent )
		    printf("Wierd expression.\n");
		goto skiplook;
	    }
	}

	repeat = rpt - 1;	/* re-execute this many times */

doitagain:
	r_low = 0;
	r_high = vocsize;

	if( dorange )
	    findrange(buf, voc);
	if( !brute )
	    numgrams = findgrams(buf, grams);
	else
	    numgrams = 0;

	if( numgrams == 0 || r_high-r_low <= threshold )
	{
	    if( repeat == 0 )
	    {
		brutes++;
		prevblock = -1;
	    }
	    for( i=bfac*r_low ; i<bfac*r_high && i<vocsize ; i++ )
	    {
		    /* Simulate searching for next string */
		    /* Doesn't appear to affect times */
		for( j=0 ; voc[i][j] != '\0' ; j++ )
		    ;
		if( repeat == 0 && prevblock != diskblocks[i] )
		{
		    prevblock = diskblocks[i];
		    blockacc++;
		}
#ifdef SOLARIS
#ifdef SLOWMATCH
		for( scnt=ss=0 ; scnt<SLOWDOWN ; scnt++ )
		    ss += ( regex(pat, voc[i]) != NULL ) ? 1 : 0 ;
#endif /* SLOWMATCH */
		if( regex(pat, voc[i]) != NULL )
#else /* SOLARIS */
#ifdef SLOWMATCH
		for( scnt=ss=0 ; scnt<SLOWDOWN ; scnt++ )
		    ss += ( re_exec(voc[i]) ) ? 1 : 0 ;
#endif /* SLOWMATCH */
		if( re_exec(voc[i]) )
#endif /* SOLARIS */
		{
		    if( repeat == 0 )
		    {
			tc++;
			if( !timeonly && !silent )
			    printf("%s\n", voc[i]);
		    }
		}
		else if( repeat == 0 )
		    fc++;
	    }
	}
	else
	{
	    if( skip )
		numwds = findsmatches(grams, numgrams, wdnums); 
	    else
	    	numwds = findmatches(grams, numgrams, wdnums);
	    if( repeat == 0 )
		prevblock = -1;

	    for( j=0 ; j<numwds ; j++ )
	    {
		off = bfac*wdnums[j];
		for( k=0 ; k<bfac && off+k<vocsize ; k++ )
		{
		    if( k > 0 ) /* Simulate searching for next string */
			/* Doesn't appear to affect times */
			for( i=0 ; voc[off+k][i] != '\0' ; i++ )
			    ;
		    if( repeat == 0 && prevblock != diskblocks[off+k] )
		    {
			prevblock = diskblocks[off+k];
			blockacc++;
		    }
#ifdef SOLARIS
#ifdef SLOWMATCH
		    for( scnt=ss=0 ; scnt<SLOWDOWN ; scnt++ )
			ss += ( regex(pat, voc[off+k]) != NULL ) ? 1 : 0 ;
#endif /* SLOWMATCH */
		    if( regex(pat, voc[off+k]) != NULL )
#else /* SOLARIS */
#ifdef SLOWMATCH
		    for( scnt=ss=0 ; scnt<SLOWDOWN ; scnt++ )
			ss += ( re_exec(voc[off+k]) ) ? 1 : 0 ;
#endif /* SLOWMATCH */
		    if( re_exec(voc[off+k]) )
#endif /* SOLARIS */
		    {
			if( repeat == 0 )
			{
			    tc++;
			    if( !timeonly && !silent )
				printf("%s\n", voc[off+k]);
			}
		    }
		    else if( repeat == 0 )
			fc++;
		}
	    }
	}

	if( repeat > 0 )
	{
	    repeat--;
	    goto doitagain;
	}

skiplook:
#if !BRIEF
	if( !silent )
	    printf("%d true matches, %d false matches.\n", tc, fc);
#endif
	ttc += tc; tfc += fc;

	if( !silent )
	    printf("vf> ");
    }
#ifdef SOLARIS
    (void) times(&tms_now);
    milli = 1000.0 * ( ( tms_now.tms_utime + tms_now.tms_stime )
	    - ( tms_was.tms_utime + tms_was.tms_stime ) ) /
	    ( (double) CLK_TCK*qcnt*rpt );
#else /* SOLARIS */
    (void) getrusage(RUSAGE_SELF, &ru_now);
    milli = ( 1000.0*( ru_now.ru_utime.tv_sec - ru_was.ru_utime.tv_sec )
	  + 0.001*( ru_now.ru_utime.tv_usec-ru_was.ru_utime.tv_usec )
	)/( qcnt*rpt );
#endif /* SOLARIS */

    printf("#N,blk\tng\tnumq\trng,skp\tthresh\tinvlsts\tcpu\ttm/fm/blocks\tlinear  #vfind\n");
    printf("%d %d\t%d\t%d\t%c %c\t%d\t%1.3f\t%1.3f\t%1.2f/%1.2f/%1.2f %d  #vfind\n",
	GRAMLEN, bfac, ng, qcnt,  dorange?'y':'n', skip?'y':'n',
	origthresh, ilproc/(double)qcnt, milli,
	ttc/(double)qcnt, tfc/(double)qcnt, blockacc/(double)qcnt, brutes);

    /* freemem(); */

    exit(0);
}


/* Restore parameters of index and vocabulary */
void readparams(char *paramindex)
{
    FILE        *fp, *fopen();
    int		tmp;
	
    if( ( fp = fopen(paramindex, "r") ) == NULL )
    {
	fprintf(stderr, "Can't open %s for reading.\n", paramindex);
	exit(1);
    }
    fread(&vocsize, sizeof(int), 1, fp);
    fread(&noofgrams, sizeof(int), 1, fp);
    fread(&glen, sizeof(int), 1, fp);
    if( glen != GRAMLEN )
    {
	fprintf(stderr, "Need grams of length %d, %s is grams of length %d.\n",
			GRAMLEN, paramindex, glen);
	exit(1);
    }
    fread(&bfac, sizeof(int), 1, fp);
    fread(&dophonetic, sizeof(int), 1, fp);
    if( origthresh == -1 )
	fread(&origthresh, sizeof(int), 1, fp);
    else
	fread(&tmp, sizeof(int), 1, fp);	/* ignore threshold */
    fclose(fp);
}


/* Load vocabulary */
void
loadvoc(char *vocab, char **voc)
{
    char	buf[BUFLEN];
    int		i, j, len;
    int		numc = 0;
    FILE	*fp, *fopen();

    if( ( fp = fopen(vocab, "r") ) == NULL )
    {
	fprintf(stderr, "Can't open %s for reading.\n", vocab);
	exit(1);
    }

    for( i=0 ; i<=vocsize && fgets(buf, BUFLEN, fp) != NULL ; i++ )
    {
	len = strlen(buf);
	numc += len + 1;
	for( j=0 ; j<len-1 ; j++ )
	    if( !validchar(buf[j]) )
	    {
		fprintf(stderr, "Bad vocab entry: %s\n", buf);
		exit(1);
	    }
	buf[len-1] = '\0';		
	voc[i] = malloc(len);
	diskblocks[i] = ( numc >> DISKBLOCK );
	strcpy(voc[i], buf);
    }
    fclose(fp);

    if( i > vocsize )
    {
	fprintf(stderr, "Vocabulary too big.\n");
	exit(1);
    }
}


/* Load vocabulary index; sets bfac */
int
loadindex(char *vocindex)
{
    char	gram[GRAMLEN];
    int		i;
    HNODE	*thenode;
    BITVEC	*vp;
    FILE	*fp, *fopen();

    if( ( fp = fopen(vocindex, "r") ) == NULL )
    {
	fprintf(stderr, "Can't open %s for reading.\n", vocindex);
	exit(1);
    }
    for( i=0 ;; i++ )
    {
	if( fread(gram, sizeof(char), GRAMLEN, fp) < GRAMLEN )
	    break;
	thenode = hinsert(gram);
	vp = gethvec(thenode);
	if( readvec(vp, fp) < 0 )
	{
	    fprintf(stderr, "File %s abnormally terminated.\n", vocindex); 
	    exit(1);
	}
    }
    fclose(fp);
    return(i);
}


/* Load vocabulary index; sets bfac; used when skip is activated */
int
loadsindex(char *vocindex)
{
    char	gram[GRAMLEN];
    int         i;
    SNODE       *thenode;
    BITVEC      *vp,*sp,*sap;
    FILE        *fp, *fopen();

    if( ( fp = fopen(vocindex, "r") ) == NULL )
    {
	fprintf(stderr, "Can't open %s for reading.\n", vocindex);
	exit(1);
    }
    for( i=0 ;; i++ )
    {
	if( fread(gram, sizeof(char), GRAMLEN, fp) < GRAMLEN )
	    break;

	thenode = hsinsert(gram);
	vp = getshvec(thenode);
	if( readvec(vp, fp) < 0 )
	{
	    fprintf(stderr, "File %s abnormally terminated.\n", vocindex);
	    exit(1);
	}
	sp = getspvec(thenode);
	if( readvec(sp, fp) < 0 )
	{
	    fprintf(stderr, "File %s abnormally terminated.\n", vocindex);
	    exit(1);
	}
	sap = getsapvec(thenode);
	if( readvec(sap, fp) < 0 )
	{
	    fprintf(stderr, "File %s abnormally terminated.\n", vocindex);
	    exit(1);
	}
    }
    fclose(fp);
    return(i);
}


/* Set r_low, r_high with binary search on voc */
void
findrange(char *b, char *voc[])
{
    int		len = 0;			/* amt of string to compare */
    int		lb, ub;				/* all matches in this range */
    int		top, bot, mid;			/* for binary search */
    int		v1, v2, match, found;
    char	buf[BUFLEN];

    if( b[0] == '^' )
    {
	b++;
	for( ; validchar(b[len]) ; len++ )
	    buf[len] = b[len];
	buf[len] = '\0';
	if( b[len] == '$' && b[len+1] == '\0' )
		/* this is an exact match; do strcmp, not strncmp */
	    len = BUFLEN;
    }
    if( len <= 0 )
	return;

	/* find any string starting with prefix */
    lb = bot = r_low, ub = top = r_high - 1;
    for( found=false ; !found && bot <= top ; )
    {
	lb = bot, ub = top;
	mid = ( top + bot ) / 2;
	v1 = strncmp(buf, voc[mid], len);
	if( v1 < 0 )
	    top = mid - 1;
	else if( v1 > 0 )
	    bot = mid + 1;
	else
	    found = true;
    }
    match = mid;

    if( found )
    {
	    /* find first string starting with prefix */
	bot = lb, top = match;
	for( found=false ; !found && bot <= top ; )
	{
	    mid = ( top + bot ) / 2;
	    v1 = strncmp(buf, voc[mid], len);
	    v2 = ( mid>=r_low && mid>0 ) ? strncmp(buf, voc[mid-1], len) : -1;
	    if( v1 > 0 )
		bot = mid + 1;
	    else if( v2 == 0 )
		top = mid - 1;
	    else
		found = true;
	}
	r_low = mid;

	    /* find last string starting with prefix */
	bot = match, top = ub;
	for( found=false ; !found && bot <= top ; )
	{
	    mid = ( top + bot ) / 2;
	    v1 = strncmp(buf, voc[mid], len);
	    v2 = ( mid<r_high && mid<vocsize-1 ) ?
		strncmp(buf, voc[mid+1], len) : 1;
	    if( v1 == 0 )
		bot = mid + 1;
	    else if( v2 < 0 )
		top = mid - 1;
	    else
		found = true;
	}
	r_high = mid + 1;
    }
    else
	r_low = mid, r_high = mid;

    r_low = r_low/bfac;
    r_high = 1 + r_high/bfac;

/*printf("%d %d %s-%s %s-%s\n", r_low, r_high, voc[r_low-1], voc[r_low], voc[r_high-1], voc[r_high]);*/
}


/* Search buf for all potential grams, and returns the number found */
/* Handles ^, $, ., and * */
int
findgrams(char *buf, char grams[][GRAMLEN])
{
    int		i, j, k, num, len;
    int		ok;				/* have a valid gram */
    int		isnew;				/* have a new gram */
    int		top;				/* last possible alpha char */

    len = strlen(buf);
    num = 0; i = 0; top = len - 1;

    if( dorange && buf[0] == '^' )
    {
	for( i++ ; validchar(buf[i]) ; i++ )
	    ;
	if( buf[i] == '$' && i==top )
	    i++;
    }

	/* Look for character sequence grams */
    for( ; i<top+2-GRAMLEN ; i++ )
    {
	for( ok=true, j=0 ; ok && j<GRAMLEN ; j++ )
	    if( !validchar(buf[i+j])
		    && !( GRAMLEN > 1 && i+j==0 && buf[i+j]=='^' )
		    && !( GRAMLEN > 1 && i+j==top && buf[i+j]=='$' ) )
		ok = false;
	if( ok && buf[i+j] != '*' )	/* have a valid gram */
	{
		/* Check that gram isn't already used */
	    for( isnew=true, k=0 ; k < num ; k++ )
		if( strncmp(grams[k], buf+i, GRAMLEN) == 0 )
		    isnew = false;
	    if( isnew )
	    {
	        strncpy(grams[num], buf+i, GRAMLEN);
		num++;
	    }
	}
    }

    return(num);
}


/* Find all words matching grams; returns the number of matches */
int
findmatches(char grams[][GRAMLEN], int numgrams, int *wdnums)
{
    BITVEC	*vp, *vecs[BUFLEN];
    HNODE	*thenode;
    int		i, j, k, ret, numwds, cmpit(BITVEC **,BITVEC **);

    for( i=0 ; i<numgrams ; i++ ) /* for each gram */
    {
	thenode = hfetch(grams[i]);

	if( thenode == NULL ) /* There won't be any matching strings */
	    return(0);
	vecs[i] = gethvec(thenode);
    }

	/* Sort them on increasing length -- want to process shortest first */
    qsort(vecs, numgrams, sizeof(BITVEC *), cmpit);

	/* Process first vector */
    ilproc++;
    resetvec(vp=vecs[0]);
    for( ret=getrunlen(vp) ; ret>=0 && ret<r_low ; ret=getrunlen(vp) )
	;
    for( j=0 ; ret>=0 && ret<r_high && j<QUERYMAX ; j++, ret=getrunlen(vp) )
	wdnums[j] = ret;    /* Store indexes of first gram(vector)*/

    if( j >= QUERYMAX )
    {
	fprintf(stderr, "Too many matches!\n");
	exit(1);
    }
    numwds = j;
    if( numwds <= threshold ) /* it's more economical to check for f.m. */
	return(numwds);

    for( i=1 ; i<numgrams ; i++ ) /* for each remaining vector */
		/* Merge existing list with new values as they
		   emerge, only keeping values from both */
    {
	ilproc++;
	resetvec(vp=vecs[i]);
		/* j is top of merged list, k is top of prev list */
	for( j=k=0 ; ( ret = getrunlen(vp) ) >= 0 && k < numwds ; )
	{
	    while( ret > wdnums[k] && k < numwds )
		k++;
	    if( ret == wdnums[k] )
		wdnums[j] = ret, j++, k++;
	}
	numwds = j;
	if( numwds <= threshold ) /* it's more economical to check for f.m. */
	    return(numwds);
    }

    return(numwds);
}



/* Find all words matching grams; returns the number of matches */
/* Used when skip is activated */
int
findsmatches(char grams[][GRAMLEN], int numgrams, int *wdnums)
{
    BITVEC	*vp, *vecs[BUFLEN],*svecs[BUFLEN],*saddrvecs[BUFLEN];
    SNODE	*thenode;
    int		i, j, k, ret, numwds,cmpits(SNODE **v1,SNODE **v2);
    BITVEC	*sp, *sap;
    int 	present=false;
    int 	spret,sapret,spprev,sapprev;
    int 	p;
    SNODE 	*node[BUFLEN];


    for( i=0 ; i<numgrams ; i++ )
    {
	thenode = hsfetch(grams[i]);
	node[i] = thenode;
	if( thenode == NULL )
	    return(0);
    }

    /* Sort them on increasing length -- want to process shortest first */
    qsort(node, numgrams, sizeof(SNODE *), cmpits);


    for( i=0 ; i<numgrams ; i++ ) /* for each gram */
    {
	vecs[i] = getshvec(node[i]);
	svecs[i] = getspvec(node[i]);
	saddrvecs[i]=getsapvec(node[i]);
	/*free(node[i]);*/
    }
	/*free(node);*/
    /* must free the node vector here */

    /* Process first vector : decode the first vector */
    ilproc++;
    resetvec(vp=vecs[0]);
    for( ret=getrunlen(vp) ; ret>=0 && ret<r_low ; ret=getrunlen(vp) )
	;
    for( j=0 ; ret>=0 && ret<r_high && j<QUERYMAX ; j++, ret=getrunlen(vp) )
	wdnums[j] = ret;    /* Store indexes of first gram(vector)*/

    if( j >= QUERYMAX )
    {
	fprintf(stderr, "Too many matches!\n");
	exit(1);
    }

    numwds = j;
    if( numwds <= threshold )
	return(numwds);

    for( i=1 ; i<numgrams ; i++ ) /* for each remaining vector */
	        /* Merge existing list with new values as they
	           emerge, only keeping common values*/
    {
	ilproc++;
	resetvec(vp = vecs[i]);
	resetvec(sp = svecs[i]);
	resetvec(sap = saddrvecs[i]);

	ret=getrunlen(vp);
	spprev=ret;		/* start of vector */
	sapprev=0;
	spret=getrunlen(sp);	/* start of second block */
	sapret=getrunlen(sap);

	for( j=k=0 ; k < numwds && spret >= 0 ; k++)
	{
	    present=false;
	    if( wdnums[k] == ret )
		present = true;
	    else /* skipping */
	        while( wdnums[k]>=spret && spret>=0 ) /* not in current block */
		{
		    if( spret == wdnums[k] )
			present = true;
		    ret = spprev = spret;
		    sapprev = sapret;
		    spret = getrunlen(sp);
		    sapret = getrunlen(sap);
		 }

	        /* now search the inverted list */
	    if ( !present && spret>=0 )
	    {
		if( spprev > vp->last ) /* have skipped, fix up vector */
		{
		    vp->pos = (int) (sapprev/8);
		    vp->bit = sapprev%8;
		    vp->cur = vp->vector[vp->pos];
		    vp->last = spprev;
		}
		while( ret<spret && ret<=wdnums[k] && ret>=0 )
	      	{
		    if( ret==wdnums[k] )
			present = true;
		    ret = getrunlen(vp);
	      	}
	    }
	    if( present )
	    {
		wdnums[j] = wdnums[k];
		j++;
	    }
	}

	numwds = j;
	if( numwds <= threshold )
	    return(numwds);
    }
    return(numwds);
}


/* Free up all memory allocated */
void
freemem(void)
{
    int i;

    for(i=0 ; i < vocsize ; i++)
        free(voc[i]);
    free(voc);
    if (skip)
        sfreehash();
    else
        freehash();
}


int
cmpit(BITVEC **v1, BITVEC **v2)
{
    return( veclen(*v1) - veclen(*v2) );
}


int
cmpits(SNODE **v1,SNODE **v2)
{
    return( (*v1)->vec.len - (*v2)->vec.len );
}
