/* VRANK pattern matching suite
 * Copyright (C) 1996 to Justin Zobel
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * See Copyright.txt for further details. */


/* BUG: slow, needs to have an array of strlens; see BLAH */

/* Query front end for ranking words in vocabulary. */
/* Usage is vrank [-e] [-c] [-s] [-rpt num]
		    [-a num] [-i index] [-m mode] [-n num] [-v vocab]
    -c: produce output for combining;
    -e: produce output for evaluation;
    -s: silent; show final stats only;
    -rpt num: do each query num times over, to improve accuracy of timing
    -a num: override default number of accumulators (NUMACCUM)
    -i index: override default index (INDSUF)
    -m mode: override default ranking (QGRAM)
    -n num: override default number of answers (NUMANS)
    -v vocab: override default vocabulary (VOCAB)
*/
/* Can't use skipped indexes */

/* If HASHED set, uses a limited accumulator scheme like that in FTR.
   Otherwise use an array of accumulators. */
#ifndef HASHED
#define HASHED false
#endif /* HASHED */

#include "def.h"

#if HASHED
#define NUMACCUM	1000			/* how many accumulators */
#endif /* HASHED */

#define HEAPFAC		100			/* multiplier on numans */
#define HEAPFRAC	10			/* divisor on vocsize */


typedef struct matchwordrec
{
    int		num;				/* num of matching word */
    float	wgt;				/* of word */
#if HASHED
    int		next;				/* next wd with same hval */
#endif /* HASHED */
} MATCHWORD;

int	bfac;					/* blocking factor */
int	dophonetic;				/* whether to use phonetic */
MATCHWORD *wds;					/* matching words */
int	*heap;					/* of best matches */
int	heapsize;
char	**voc;					/* the vocabulary */
int	vocsize;
int	numans = NUMANS;			/* number of answers */
int	mode = QGRAM;				/* type of ranking */
int	noofgrams;
int	glen;

#if HASHED
int	numaccum = NUMACCUM;			/* number of accumulators */

static void createhtable();
static void inithtable();
static int addaccum(int);
#endif /* HASHED */

static void readparams(char *);
static void loadvoc(char *, char **);
static int findgrams(char *,char [BUFLEN][GRAMLEN],int); /* MAX */
static int loadindex(char *);
static int cmpvec(BITVEC **,BITVEC **);		/* for sort */
static int cmpmatch(int *,int *);	/* for sort */
static int cmpnum(MATCHWORD *,MATCHWORD *);	/* for sort */
static char *massage_query(char *,int *,char *);
static void findranks(char [BUFLEN][GRAMLEN],int,char *);	/* MAX */
static void buildheap(int);
static void heapify(int,int);
static void freemem(void);

main(int argc, char *argv[])
{
    int		off, ng, qlen, qcnt;
    int		i, j, k;
    int		getrusage(), ftime();
    int		silent = false;			/* final status only */
    int		eval = false;			/* output for ranking */
    int		comb = false;			/* output for combining */
    int		manytimes = false;		/* execute each q rpt times */
    int		numgrams;			/* numbers of things */
    int		tmatch, match;			/* numbers of matches */
    int		rpt = 1;			/* tot num of iterations */
    int		repeat;				/* number of iterations left */
    char	buf[BUFLEN], grams[BUFLEN][GRAMLEN], *rindex(); /* MAX */
    char        *paramsuf = PARAMSUF, paramindex[BUFLEN];
    char	*ptr, buf2[BUFLEN];
    char	origq[BUFLEN], query[BUFLEN];	/* query string, in alnum */
    char	*vocab = VOCAB, *indsuf = INDSUF, vocindex[BUFLEN];
    double	milli;
#ifdef SOLARIS
    struct tms	tms_was, tms_now;
#else /* SOLARIS */
    struct rusage ru_was, ru_now;
#endif /* SOLARIS */

#if DIRECT
    fprintf(stderr, "Querying not implemented for direct indexes.\n");
    exit(1);
#endif /* DIRECT */

restart:
#if HASHED
    if( argc >= 3 && strcmp(argv[1], "-a") == 0 )
    {
	numaccum = atoi(argv[2]);
	argc -= 2, argv += 2;
	goto restart;
    }
#endif /* HASHED */
    if( argc >= 3 && strcmp(argv[1], "-i") == 0 )
    {
	indsuf = argv[2];
	argc -= 2, argv += 2;
	goto restart;
    }
    if( argc >= 3 && strcmp(argv[1], "-m") == 0 )
    {
	mode = atoi(argv[2]);
	if( ranktype(mode) == NULL )
	{
	    fprintf(stderr, "Rank mode %d unknown.\n", mode);
	    exit(1);
	}
	argc -= 2, argv += 2;
	goto restart;
    }
    if( argc >= 3 && strcmp(argv[1], "-n") == 0 )
    {
	numans = atoi(argv[2]);
	argc -= 2, argv += 2;
	goto restart;
    }
    if( argc >= 3 && strcmp(argv[1], "-v") == 0 )
    {
	vocab = argv[2];
	argc -= 2, argv += 2;
	goto restart;
    }
    if( argc >= 3 && strcmp(argv[1], "-rpt") == 0 )
    {
	rpt = atoi(argv[2]);
	manytimes = true;
	argc -= 2, argv += 2;
	goto restart;
    }
    if( argc >= 2 && strcmp(argv[1], "-c") == 0 )
    {
	comb = true;
	eval = false;
	argc--, argv++;
	goto restart;
    }
    if( argc >= 2 && strcmp(argv[1], "-e") == 0 )
    {
	eval = true;
	comb = false;
	argc--, argv++;
	goto restart;
    }
    if( argc >= 2 && strcmp(argv[1], "-s") == 0 )
    {
	silent = true;
	argc--, argv++;
	goto restart;
    }
    if( argc >= 2 )
    {
	fprintf(stderr, "Usage: vrank [-c] [-e] [-s] [-rpt num] [-m mode]\n");
#if HASHED
	fprintf(stderr, "             [-a num] [-i index] [-n num] [-v voc]\n");
#else /* HASHED */
	fprintf(stderr, "             [-i index] [-n num] [-v voc]\n");
#endif /* HASHED */
	exit(1);
    }

    sprintf(vocindex, "%s.%s", vocab, indsuf);
    sprintf(paramindex, "%s.%s.%s", vocab, indsuf, paramsuf);
    readparams(paramindex);

    heapsize = HEAPFAC*numans;
    if( heapsize>vocsize )
	heapsize = vocsize;
    else if( heapsize<vocsize/HEAPFRAC )
	heapsize = vocsize/HEAPFRAC;
    heap = (int *) malloc(heapsize * sizeof(int));
#if HASHED
    if( numaccum < heapsize )
	numaccum = heapsize;
    wds = (MATCHWORD *) malloc(numaccum * sizeof(MATCHWORD));
    createhtable();
#else /* HASHED */
    wds = (MATCHWORD *) malloc(vocsize * sizeof(MATCHWORD));
#endif /*HASHED */

#ifdef SOLARIS
    (void) times(&tms_was);
#else /* SOLARIS */
    (void) getrusage(RUSAGE_SELF, &ru_was);
#endif /* SOLARIS */

    voc = (char **) malloc(vocsize*sizeof(char*));
    inithash(noofgrams);
    loadvoc(vocab, voc);
    ng = loadindex(vocindex);	/* sets bfac, dophonetic */

#ifdef SOLARIS
    (void) times(&tms_now);
    milli = 1000.0 * ( ( tms_now.tms_utime + tms_now.tms_stime )
	    - ( tms_was.tms_utime + tms_was.tms_stime ) ) / (double) CLK_TCK;
#else /* SOLARIS */
    (void) getrusage(RUSAGE_SELF, &ru_now);
    milli = 1000.0 * ( ru_now.ru_utime.tv_sec - ru_was.ru_utime.tv_sec )
	    + 0.001*( ru_now.ru_utime.tv_usec - ru_was.ru_utime.tv_usec );
#endif /* SOLARIS */

    if( bfac > 1 )
    {
	fprintf(stderr, "Error: words are grouped into blocks.\n");
	exit(1);
    }

    if( !comb && !eval && !silent )
	printf("Phonetic search type is %s.\n", phonetictype(dophonetic));
    if( dophonetic == PHONIX )
	loadphontable();

#ifdef NEVERDEF
    if( !eval && !comb ) 
	printf("Startup took %1.3f cpu sec.\n", milli/1000.0);
#endif /* NEVERDEF */

	/* Get a string, parse into grams, merge ptr lists of
	   grams, print out all matching words. */
	/* Naive: doesn't go for shortest ptr list, etc. */
    if( !silent && !eval && !comb )
	printf("vr> ");
#ifdef SOLARIS
    (void) times(&tms_was);
#else /* SOLARIS */
    (void) getrusage(RUSAGE_SELF, &ru_was);
#endif /* SOLARIS */
    for( tmatch=qcnt=0 ; ( ptr = gets(buf) ) != NULL ; qcnt++ )
    {
	match = 0;

	for( k=j=0 ; buf[j]!='\0' ; j++ )
	    if( validchar(buf[j]) )
	    {
		origq[k] = buf[j];
		k++;
	    }
	origq[k] = '\0';
	qlen = j;

	/* if( dophonetic ) */
	    ptr = massage_query(buf, &qlen, buf2);
	for( k=j=0 ; ptr[j]!='\0' ; j++ )
	    if( validchar(ptr[j]) )
	    {
		query[k] = ptr[j];
		k++;
	    }
	query[k] = '\0';
	qlen = k;

	if( qlen == 0 )
	    goto skiplook;

	repeat = rpt - 1;	/* re-execute `rpt' many times */

doitagain:
	numgrams = findgrams(ptr, grams, GRAMLEN);

	if( numgrams == 0 )
	{
	    if( !eval && !comb )
		fprintf(stderr, "Can't proceed -- no grams given\n");
	}
	else
	{
	    /* if( mode == EDITRAW )
		findranks(grams, numgrams, origq);
	    else */
		findranks(grams, numgrams, query);
	    if( repeat == 0 && eval )
		printf("%s", origq);
	    /* probable BUG --should stop when wgt <=0 */
	    for( j=0 ; j<numans && wds[heap[j]].num != -1 ; j++ )
	    {
		off = wds[heap[j]].num;
		{
		    if( repeat == 0 && wds[heap[j]].wgt>0 ) /* BUG untested */
		    {
			match++;
			if( eval )
			    printf(" %s", voc[off]);
			else if( comb )
			    printf("%s %1.3f\n", voc[off], wds[heap[j]].wgt);
			else if( !silent )
			    printf("%-15s [%1.2f]\n",voc[off],wds[heap[j]].wgt);
		    }
		}
	    }
	    if( repeat == 0 && eval )
		printf("\n");
	}

	if( repeat > 0 )
	{
	    repeat--;
	    goto doitagain;
	}

skiplook:
#ifdef NEVERDEF
	if( !silent && !eval && !comb )
	    printf("%d matches.\n", match);
#endif /* NEVERDEF */
	tmatch += match;

	if( !silent && !eval && !comb )
	    printf("vr> ");
    }
#ifdef SOLARIS
    (void) times(&tms_now);
    milli = 1000.0 * ( ( tms_now.tms_utime + tms_now.tms_stime )
	    - ( tms_was.tms_utime + tms_was.tms_stime ) ) /
	    ( (double) CLK_TCK*qcnt*rpt );
#else /* SOLARIS */
    (void) getrusage(RUSAGE_SELF, &ru_now);
    milli = ( 1000.0*( ru_now.ru_utime.tv_sec - ru_was.ru_utime.tv_sec )
	  + 0.001*( ru_now.ru_utime.tv_usec-ru_was.ru_utime.tv_usec )
	)/( qcnt*rpt );
#endif /* SOLARIS */

    if( !eval && !comb )
    {
	fprintf(stderr,"#N\tng\tnumq\tblock\tphonet.\trank\tcpu\tmatch\taccum  #vrank\n");
#if HASHED
	fprintf(stderr,"%d\t%d\t%d\t%d\t%s\t%s\t%1.3f\t%d\t%d  #vrank\n",
	    GRAMLEN, ng, qcnt, bfac, phonetictype(dophonetic), ranktype(mode),
	    milli,
	    tmatch, numaccum);
#else /* HASHED */
	fprintf(stderr,"%d\t%d\t%d\t%d\t%s\t%s\t%1.3f\t%d\t%d  #vrank\n",
	    GRAMLEN, ng, qcnt, bfac, phonetictype(dophonetic), ranktype(mode),
	    milli,
	    tmatch, vocsize);
#endif /* HASHED */
    }
    
    /* freemem(); */

    exit(0);
}


void readparams(char *paramindex)
{
    FILE        *fp, *fopen();
    int		tmp;
 
    if( ( fp = fopen(paramindex, "r") ) == NULL )
    {
	fprintf(stderr, "Can't open %s for reading.\n", paramindex);
	exit(1);
    }
    fread(&vocsize, sizeof(int), 1, fp);
    fread(&noofgrams, sizeof(int), 1, fp);
    fread(&glen, sizeof(int), 1, fp);
    if( glen != GRAMLEN )
    {
	fprintf(stderr, "Need grams of length %d, %s is grams of length %d.\n",
			GRAMLEN, paramindex, glen);
	exit(1);
    }
    fread(&bfac, sizeof(int), 1, fp);
    fread(&dophonetic, sizeof(int), 1, fp);
    fread(&tmp, sizeof(int), 1, fp);
    fclose(fp);
}


/* Load vocabulary */
void
loadvoc(char *vocab, char **voc)
{
    char	buf[BUFLEN];
    int		i, j, len;
    FILE	*fp, *fopen();

    if( ( fp = fopen(vocab, "r") ) == NULL )
    {
	fprintf(stderr, "Can't open %s for reading.\n", vocab);
	exit(1);
    }

    for( i=0 ; i<=vocsize && fgets(buf, BUFLEN, fp) != NULL ; i++ )
    {
	len = strlen(buf);
	for( j=0 ; j<len-1 ; j++ )
	    if( !validchar(buf[j]) )
	    {
		fprintf(stderr, "Bad vocab entry: %s\n", buf);
		exit(1);
	    }
	buf[len-1] = '\0';		/* overwrite '\n' */
	voc[i] = malloc(len * sizeof(char));
	strcpy(voc[i], buf);
    }
    fclose(fp);

    if( i > vocsize )
    {
	fprintf(stderr, "Vocabulary too big.\n");
	exit(1);
    }
}


/* Load vocabulary index; sets bfac */
int
loadindex(char *vocindex)
{
    char	gram[GRAMLEN];
    int		i;
    HNODE	*thenode;
    BITVEC	*vp;
    FILE	*fp, *fopen();

    if( ( fp = fopen(vocindex, "r") ) == NULL )
    {
	fprintf(stderr, "Can't open %s for reading.\n", vocindex);
	exit(1);
    }
    for( i=0 ;; i++ )
    {
	if( fread(gram, sizeof(char), GRAMLEN, fp) < GRAMLEN )
	    break;
	thenode = hinsert(gram);
	vp = gethvec(thenode);
	if( readvec(vp, fp) < 0 )
	{
	    fprintf(stderr, "File %s abnormally terminated.\n", vocindex); 
	    exit(1);
	}
    }
    fclose(fp);
    return(i);
}


/* Search buf for all potential grams, and returns the number found.
   Grams are returned in appearance order */
/* Handles ^, $, ., and * */
int
findgrams(char *buf, char grams[BUFLEN][GRAMLEN], int gramlen) /* MAX */
{
    int		i, j, k, num, len;
    int		ok;				/* have a valid gram */
    int		isnew;				/* have a new gram */
    int		top;				/* last possible alpha char */

    len = strlen(buf);
    num = 0; i = 0; top = len - 1;

	/* Look for character sequence grams */
    for( ; i<top+2-gramlen ; i++ )
    {
	for( ok=true, j=0 ; ok && j<gramlen ; j++ )
	    if( !validchar(buf[i+j])
		    && !( gramlen > 1 && i+j==0 && buf[i+j]=='^' )
		    && !( gramlen > 1 && i+j==top && buf[i+j]=='$' ) )
		ok = false;
	if( ok && buf[i+j] != '*' )	/* have a valid gram */
	{
		/* Check that gram isn't already used */
	    for( isnew=true, k=0 ; k < num ; k++ )
		if( strncmp(grams[k], buf+i, gramlen) == 0 )
		    isnew = false;
	    if( isnew )
	    {
	        strncpy(grams[num], buf+i, gramlen);
		num++;
	    }
	}
    }

    return(num);
}


/* Find all words matching grams; returns the number of matches */
void
findranks(char grams[BUFLEN][GRAMLEN], int numgrams, char *query) /* MAX */
{
    BITVEC	*vp, *vecs[BUFLEN];		/* bitvectors */
    HNODE	*thenode;			/* for curr. gram */
    int		i, j, ret;
    char	buf2[BUFLEN], *ptr;
    char	gram[GRAMLEN];
    int		lim;				/* of words */
    int		qlen;
    int		curr = 0;			/* ptr to wds */
    float	maxwgt, minwgt;			/* max, min in heap */
    int		lendiff;
#if HASHED
    int		runout = false;			/* whether slots left */
#endif /* HASHED */

#if HASHED
    lim = numaccum;
    inithtable();
#else /* HASHED */
    lim = vocsize;
#endif /* HASHED */

	/* Get all vectors */
    for( j=i=0 ; i<numgrams ; i++ ) /* for each gram */
    {
	strncpy(gram, grams[i], GRAMLEN);
	thenode = hfetch(gram);
	if( thenode != NULL ) /* there are matching strings */
	{
	    vecs[j] = gethvec(thenode);
	    j++;
	}
    }
    numgrams = j;		/* forget about unused grams */

    /* order doesn't matter */
#ifdef NEVERDEF
    if( mode != PRIORITY )
	    /* Sort on increasing length -- want to process shortest first */
	sort(vecs, numgrams, sizeof(BITVEC *), cmpvec);
#endif /* NEVERDEF */

    for( i=0 ; i<lim ; i++ )
    {
#if HASHED
	wds[i].num = -1;
#else /* HASHED */
	wds[i].num = i;
#endif /* HASHED */
	wds[i].wgt = 0.0;
    }

    for( i=0 ; i<numgrams ; i++ ) /* process each vector */
    {
	resetvec(vp=vecs[i]);
#if HASHED
	if( curr != -1 )	/* still have space in hash table */
	{
	    while( ( ret = getrunlen(vp) ) >= 0 && ! runout )
	    {
		curr = addaccum(ret);
		if( curr == -1 )	/* hash table out of space */
		{
		    if( ! runout )	/* just ran out of space */
			sort(wds, lim, sizeof(MATCHWORD), cmpnum);
		    runout = true;
		}
		else
		{
		    if( mode == PRIORITY )
			/* give high wgt to early grams */
			/* increasing const => lower the importance of order */
			wds[curr].wgt += 1.0/( 2.0 + i );
		    else if( mode == EDIT3TAPER || mode == EDITEXTAPER )
			wds[curr].wgt += 2*numgrams - i;
		    else
			wds[curr].wgt += 1.0;
		}
	    }
	}
	else			/* hash table out of space */
	{
	    ret = getrunlen(vp);
	    for( curr=0 ; curr<lim && ret>=0 ; curr++ )
	    {
		while( wds[curr].num>ret && ret>=0 )
		    ret = getrunlen(vp);
		if( wds[curr].num == ret )
		{
		    if( mode == PRIORITY )
			wds[curr].wgt += 1.0/( 2.0 + i );
		    else
			wds[curr].wgt += 1.0;
		}
	    }
	}
#else /* HASHED */
	while( ( ret = getrunlen(vp) ) >= 0 )
	{
	    if( mode == PRIORITY )
		wds[ret].wgt += 1.0/( 2.0 + i );
	    else
		wds[ret].wgt += 1.0;
	}
#endif /* HASHED */
    }

    /* All wgts MUST be positive at this stage */

    qlen = strlen(query);

	/* build heap of highest weights */
    lendiff = strlen(voc[i]) - qlen;
    if( lendiff<0 )
	lendiff = -lendiff;

    /* BLAH */ /* for(i=0 ;i<lim ; i++ ) wds[i].wgt /= ( 5+lendiff ); */
    if( lim<heapsize )
	heapsize = lim;
    for( i=0 ; i<heapsize ; i++ )
	heap[i] = i;
    buildheap(heapsize);
    for( ; i<lim ; i++ )
	if( wds[heap[0]].wgt < wds[i].wgt )
	{
	    heap[0] = i;
	    heapify(0, heapsize);
	}

	/* normalise these words */
    for( i=0 ; i<heapsize ; i++ )
    {
	curr = heap[i];
	if( wds[curr].num == -1 )
	    wds[curr].wgt = -1000000;
	else if( mode != PLAIN )
	{
	    if( dophonetic /* && mode != EDITRAW */ )
		/* transform the word */
	    {
		ptr = voc[wds[curr].num];
		phonetic(dophonetic, ptr, strlen(ptr), buf2);
		ptr = buf2;
	    }
	    else
		ptr = voc[wds[curr].num];

	    if( mode == PRIORITY )
		    /* increasing const => lower the importance of length */
		wds[curr].wgt /= ( 4.0 + diff(qlen, strlen(ptr)) );
	    else if( mode == QGRAM )
	    {
/* #define DUMBFORMULA 1 */
#ifdef DUMBFORMULA
/* as in Pfeifer, Poersch, Fuhr */
		wds[curr].wgt = (double)wds[curr].wgt /
		    (double)( strlen(ptr)-GRAMLEN+1 + qlen-GRAMLEN+1 - wds[curr].wgt );
#else
		wds[curr].wgt = wds[curr].wgt / (double)( strlen(ptr) + qlen );
		/* wds[curr].wgt = qlen + 2.0 * wds[curr].wgt
		    // - (qlen-GRAMLEN+1) // - (strlen(ptr)-GRAMLEN+1); */
#endif
	    }
	    else if( mode == EDIT3 /* || mode == EDITRAW */ )
	    {
		wds[curr].wgt = 1000.0 - (float) edit3(query, ptr);
#define MAXX(a,b) ( (a>b) ? a : b )
		/* wds[curr].wgt = (double)( MAXX(qlen,strlen(ptr)) - edit3(query, ptr) )/(double)( strlen(ptr) + qlen ); */
		/* wds[curr].wgt = 1000.0 - (double)edit3(query, ptr)/(double)( strlen(ptr) + qlen ); */
	    }
	    else if( mode == ONECHAR )
		wds[curr].wgt = (float) ( ( edit3(query, ptr) <= 1 ) ? 1 : 0 );
	    else if( mode == LCS )
		wds[curr].wgt = (float) lcs(query, ptr) / ( strlen(ptr) + qlen );
	    else if( mode == EDIT3TAPER )
		wds[curr].wgt = 1000.0 - (float) edit3taper(query, ptr);
	    else if( mode == EDITEX )
		wds[curr].wgt = 1000.0 - (float) edit3ex(query, ptr);
	    else if( mode == EDITEXTAPER )
		wds[curr].wgt = 1000.0 - (float) edit3extaper(query, ptr);
	    else if( mode == EDIT4 )
		wds[curr].wgt = 1000.0 - (float) edit4(query, ptr);
	}
    }

	/* sort heap */
    sort(heap, heapsize, sizeof(int), cmpmatch);

	/* Now normalise the weights in the heap */
    for( minwgt=100000000.0, maxwgt=0.0, i=0 ; i<heapsize ; i++ )
    {
	if( wds[heap[i]].wgt > maxwgt )
	    maxwgt = wds[heap[i]].wgt;
	if( wds[heap[i]].wgt < minwgt )
	    minwgt = wds[heap[i]].wgt;
    }

    for( i=0 ; i<heapsize ; i++ )
	wds[heap[i]].wgt = 100.0 * (wds[heap[i]].wgt-minwgt) / (maxwgt-minwgt);

    return;
}


int
diff(int i1, int i2)
{
    return( ( i1>i2 ) ? i1-i2 : i2-i1 );
}


#if HASHED
int	*htable;
int	numwds;

void
createhtable()
{
    htable = (int *) malloc(numaccum * 2 * sizeof(int));
}


void
inithtable()
{
    int		i;

    numwds = 0;
    for( i=0 ; i<numaccum * 2 ; i++ )
	htable[i] = -1;		/* mark node as unused */
}


/* Returns address of accumulator for wdnum, of -1 if none */
int
addaccum(int wdnum)
{
    int		hval, curr;

    hval = ( wdnum * 17 ) % numaccum;
    curr = htable[hval];
    if( curr == -1 )
    {
	if( numwds < numaccum-1 )
	{
	    htable[hval] = numwds;
	    curr = numwds;
	    wds[curr].num = wdnum;
	    wds[curr].next = -1;
	    numwds++;
	}
    }
    else
    {
	while( wds[curr].num != wdnum && wds[curr].next != -1 )
	    curr = wds[curr].next;
	if( wds[curr].num != wdnum && numwds < numaccum-1 )
	{
	    wds[curr].next = numwds;
	    curr = numwds;
	    wds[curr].num = wdnum;
	    wds[curr].next = -1;
	    numwds++;
	}
    }
    return(curr);
}
#endif /* HASHED */


/* Massage query into phonetic form, taking care of ^ and $ */
char *
massage_query(char *buf, int *qlen, char *buf2)
{
    /* int	hat, dollar;			/ true if ^, $ used in str */
	/* but are always used in string ... */
    char	*ptr = buf;

    /* hat = dollar = false; */
    if( ptr[*qlen-1] == '$' )
    {
	/* dollar = true; */
	ptr[*qlen-1] = '\0';
	(*qlen)--;
    }
    if( ptr[0] == '^' )
    {
	/* hat = true; */
	ptr++;
	(*qlen)--;
    }
    phonetic(dophonetic, ptr, *qlen, buf2);
    /* if( hat )
    { */
	strcpy(buf+1, buf2);
	ptr = buf;
	ptr[0] = '^';
    /* }
    else
	ptr = buf2; */
    *qlen = strlen(ptr);
    /* if( dollar )
    { */
	ptr[*qlen] = '$';
	(*qlen)++;
	ptr[*qlen] = '\0';
    /* } */

    return(ptr);
}


int
cmpvec(BITVEC **v1, BITVEC **v2)
{
    return( veclen(*v1) - veclen(*v2) );
}


int
cmpmatch(int *w1, int *w2)
{
    return( ( wds[*w2].wgt>wds[*w1].wgt ) ? 1 :
	    ( wds[*w2].wgt<wds[*w1].wgt ) ? -1 :
	    ( wds[*w2].num>wds[*w1].num ) ? 1 :
	    ( wds[*w2].num<wds[*w1].num ) ? -1 : 0 );
}


int
cmpnum(MATCHWORD *w1, MATCHWORD *w2)
{
    return( w1->num-w1->num );
}


void
buildheap(int hsize)
{
    int		i;

    for( i=hsize/2 ; i>=0 ; i-- )
	heapify(i, hsize);
}


void
heapify(int pos, int hsize)
{
    int		l, r, smallest, tmp;

    l = 2*pos + 1;
    r = 2*pos + 2;
    if( l<hsize && wds[heap[l]].wgt < wds[heap[pos]].wgt )
	smallest = l;
    else
	smallest = pos;
    if( r<hsize && wds[heap[r]].wgt < wds[heap[smallest]].wgt )
	smallest = r;
    if( smallest != pos )
    {
	tmp = heap[pos];
	heap[pos] = heap[smallest];
	heap[smallest] = tmp;
	heapify(smallest, hsize);
    }
}


void freemem(void)
{
    int i;
 
    for(i=0 ; i < vocsize ; i++)
        free(voc[i]);
    free(voc);
    freehash();
}
